/*
 * Copyright (c) 1992, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static  char rcsid[] =
    "@(#)$Header: io.c,v 1.6 96/07/04 19:27:10 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ppsclock.h>	/* optionally defines PPSCLOCKSTR */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#ifdef PPSCLOCKSTR
#include <sys/stropts.h>
#else
#include <sys/ioctl.h>
#endif

#include "gpstime.h"
#include "util.h"
#include "io.h"

int
gpsinit(fd)
	register int fd;
{
	struct termios termios;
#ifndef PPSCLOCKSTR
	int ldisc;
#endif

	memset((char *)&termios, 0, sizeof(termios));
	termios.c_cflag = CS8 | CREAD | CLOCAL;
	termios.c_iflag = IGNCR;
	termios.c_lflag = ICANON;
	if (cfsetispeed(&termios, B4800) < 0 ||
	    cfsetospeed(&termios, B4800) < 0) {
		fprintf(stderr, "%s: set speed failed\n", prog);
		perror("cfsetspeed");
		return (0);
	}
	if (tcsetattr(fd, TCSAFLUSH, &termios) < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("tcsetattr");
		return (0);
	}

#ifdef PPSCLOCKSTR
	/* Under SunOS, we have a ppsclock streams module */
	if (ioctl(fd, I_PUSH, PPSCLOCKSTR) < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("push ppsclock");
		return (0);
	}
#else
	/* Under BSD we have a clock line discipline */
	ldisc = CLOCKDISC;
	if (ioctl(fd, TIOCSETD, (caddr_t)&ldisc) < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("set disp");
		return (0);
	}
#endif
	return (1);
}

int
gpsready(fd, timeout)
	register int fd;
	register u_int timeout;
{
	register int n;
	struct timeval t;
	fd_set readfds;

	t.tv_sec = timeout;
	t.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);
	if ((n = select(fd + 1, &readfds, 0, 0, &t)) <= 0) {
		if (n < 0)
			perror("gpsready: select");
		return (0);
	}
	if (!FD_ISSET(fd, &readfds)) {
		fprintf(stderr, "gpsready: wrong fd\n");
		return (0);
	}
	return (1);
}

static const char char2hex[] = "0123456789ABCDEF";

/*
 * Read a string from the gps. The checksum and newline is suppressed.
 */
int
gpsgets(fd, buf, size, timeout)
	register int fd;
	register char *buf;
	register u_int size, timeout;
{
	register int n;
	register char ck, *cp;

	if (!gpsready(fd, timeout)) {
		if (debug)
			(void) fputs("<- [|]", stderr);
		return (0);
	}

	n = read(fd, buf, size - 1);
	if (n < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("gpsgets: read");
		return (0);
	}

	if (debug) {
		(void) fputs("<- \"", stderr);
		viswrite(buf, n);
		(void)fputc('"', stderr);
	}

	/* Check checksum */
	buf[n] = '\0';
	cp = strchr(buf, '*');
	if (cp == NULL) {
		if (debug)
			(void) fputs("[*]\n", stderr);
		return (0);
	}
	++cp;
	if (cp[0] == '\0' || cp[1] == '\0') {
		if (debug)
			(void) fputs("[\\0]\n", stderr);
		return (0);
	}
	ck = cksum(&buf[1], n - 5);
	if (char2hex[ck >> 4] != cp[0] || char2hex[ck & 0xf] != cp[1]) {
		if (debug)
			(void) fputs("[^]\n", stderr);
		return (0);
	}

	/* Truncate checksum */
	*--cp = '\0';

	if (debug)
		(void)fputc('\n', stderr);

	return (1);
}


void
#if __STDC__
gpssend(register int fd, char const *fmt, ...)
#else
gpssend(fd, fmt, va_alist)
	register int fd;
	const char *fmt;
	va_dcl
#endif
{
	register char *cp;
	register int n, m;
	va_list ap;
	char buf[1024];
	u_char ck;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	cp = buf;
	*cp++ = '$';
#ifdef notdef
	/* BSD is rational */
	n = vsnprintf(cp, sizeof(buf) - 1, fmt, ap);
#else
	/* SunOS sucks */
	(void)vsprintf(cp, fmt, ap);
	n = strlen(cp);
#endif
	ck = cksum(cp, n);
	cp += n;
	++n;
#ifdef notdef
	/* BSD is rational */
	n += snprintf(cp, sizeof(buf) - n - 5, "*%02X\r\n", ck);
#else
	/* SunOS sucks */
	sprintf(cp, "*%02X\r\n", ck);
	n += strlen(cp);
#endif

	m = write(fd, buf, n);
	if (m < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("gpssend: write");
		return;
	}
	if (m != n)
		fprintf(stderr, "%s: gpssend: write: %d != %d\n", prog, m, n);
	if (debug) {
		(void) fputs("-> \"", stderr);
		viswrite(buf, m);
		(void) fputs("\"\n", stderr);
	}
	va_end(ap);
}

/* Returns true if ate anything */
int
gpseatme(fd)
	register int fd;
{
	register int n;
	char buf[1024];

	if (!gpsready(fd, 0))
		return (0);

	n = read(fd, buf, sizeof(buf));
	if (n < 0) {
		fprintf(stderr, "%s: ", prog);
		perror("gpseatme: read");
		return(0);
	}
	if (n > 0 && buf[n - 1] == '\n')
		--n;
	if (debug) {
		(void)fputs("<~ \"", stderr);
		viswrite(buf, n);
		(void)fputs("\"\n", stderr);
	}
	return (1);
}
